from sql_server.pyodbc.base import *
from sql_server.pyodbc.base import DatabaseWrapper as MSSqlDatabaseWrapper
from django_pyodbc_gis.creation import MSSqlCreation
from django_pyodbc_gis.introspection import MSSqlIntrospection
from django_pyodbc_gis.operations import MSSqlOperations


class DatabaseWrapper(MSSqlDatabaseWrapper):

    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)
        self.creation = MSSqlCreation(self)
        self.ops = MSSqlOperations(self)
        self.introspection = MSSqlIntrospection(self)



from sql_server.pyodbc.creation import DatabaseCreation


class MSSqlCreation(DatabaseCreation):
    """
    This class encapsulates all backend-specific differences that pertain to
    database *creation*, such as the column types to use for particular Django
    Fields, the SQL used to create and destroy tables, and the creation and
    destruction of test databases.
    """

    def sql_indexes_for_field(self, model, f, style):
        from django.contrib.gis.db.models.fields import GeometryField
        output = super(MSSqlCreation, self).sql_indexes_for_field(model, f, style)

        if isinstance(f, GeometryField):
            qn = self.connection.ops.quote_name
            db_table = model._meta.db_table
            constraint_name = 'chk_%s_%s_%s' % \
                              (db_table, f.column, f.geom_type)

            # Geometry-type constraint:
            output.append(style.SQL_KEYWORD('ALTER TABLE ') +
                          style.SQL_TABLE(qn(db_table)) +
                          style.SQL_KEYWORD('ADD CONSTRAINT ') +
                          style.SQL_TABLE(qn(constraint_name)) +
                          style.SQL_KEYWORD(' CHECK ( ') +
                          style.SQL_FIELD(qn(f.column)) +
                          '.STGeometryType() = \'' +
                          style.SQL_KEYWORD(f.geom_type) + '\' )')

            # Spatial index; see:
            # http://technet.microsoft.com/en-us/library/bb934196.aspx
            if f.spatial_index:
                idx_name = '%s_%s_id' % (db_table, f.column)
                extent = str(f._extent)
                output.append(style.SQL_KEYWORD('CREATE SPATIAL INDEX ') +
                              style.SQL_TABLE(qn(idx_name)) +
                              style.SQL_KEYWORD(' ON ') +
                              style.SQL_TABLE(qn(db_table)) + '(' +
                              style.SQL_FIELD(qn(f.column)) + ') ' +
                              style.SQL_KEYWORD('USING GEOMETRY_AUTO_GRID ') +
                              style.SQL_KEYWORD('WITH ') + '(' +
                              style.SQL_KEYWORD('BOUNDING_BOX') + ' = ' +
                              extent + ' )')

        return output



from django.contrib.gis.db.models.sql.compiler import GeoSQLCompiler as BaseGeoSQLCompiler
from sql_server.pyodbc import compiler


SQLCompiler = compiler.SQLCompiler


class GeoSQLCompiler(BaseGeoSQLCompiler, SQLCompiler):
    pass


class SQLInsertCompiler(compiler.SQLInsertCompiler, GeoSQLCompiler):
    pass


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, GeoSQLCompiler):
    pass


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, GeoSQLCompiler):
    pass


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, GeoSQLCompiler):
    pass


class SQLDateCompiler(compiler.SQLDateCompiler, GeoSQLCompiler):
    pass


class SQLDateTimeCompiler(compiler.SQLDateTimeCompiler, GeoSQLCompiler):
    pass



import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="django-pyodbc-gis",
    version="0.0.6",
    author="Mark Hepburn",
    author_email="mark@condense.com.au",
    description=("GIS support for SQL Server, on top of django-pyodbc-azure"),
    license="BSD",
    keywords="django mssql gis",
    url="https://www.github.com/condense/django-pyodbc-gis",
    packages=['django_pyodbc_gis'],
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 1 - Planning",
        "Topic :: Database",
        "License :: OSI Approved :: BSD License",
    ],
    install_requires=[
        "django-pyodbc-azure",
    ],
)



from django.contrib.gis.gdal import OGRGeomType

from sql_server.pyodbc.introspection import DatabaseIntrospection
import re


class GeoIntrospectionError(Exception):
    pass


class MSSqlIntrospection(DatabaseIntrospection):

    def get_geometry_type(self, table_name, geo_col):
        """
        The geometry type used by SQL Server doesn't contain any
        specific type, SRID,or dimension information; that is
        associated with instances.  This makes introspection hacky and
        error prone, but presumably that also applies to the data
        integrity.
        """
        cursor = self.connection.cursor()
        try:
            field_type, field_params = None, {}
            # Look for a sample piece of data and assume that it's
            # representative.  If we can't find anything, we'll have
            # to examine the constraints:
            try:
                cursor.excute("SELECT TOP 1 [%(geo_col)].STGeometryType(), " +
                              "[%(geo_col)].STDimension(), " +
                              "[%(geo_col)].STSrid " +
                              "FROM [%(table_name)] " +
                              "WHERE [%(geo_col)] IS NOT NULL" % locals())

                row = cursor.fetchone()

                if not rows:
                    raise GeoIntrospectionError

                field_type = OGRGeomType(row[0]).django

                srid, dim = row[1:]
                if srid != 4326:
                    field_params['srid'] = srid
                if dim != 2:
                    field_params['dim'] = dim

            except GeoIntrospectionError:
                # The table is empty, so we'll have to guess based on
                # parsing the constraint information.  Unfortunately,
                # this will only give us the type, but not dimension
                # or SRID (and nothing reliably)
                cursor.execute(
                    "SELECT [CHECK_CLAUSE]"
                    "FROM [INFORMATION_SCHEMA].[CHECK_CONSTRAINTS] cc"
                    "JOIN [INFORMATION_SCHEMA].[CONSTRAINT_COLUMN_USAGE] ccu"
                    "ON ccu.[CONSTRAINT_NAME] = cc.[CONSTRAINT_NAME]"
                    "WHERE ccu.[CONSTRAINT_CATALOG] = '%s'"
                    "AND [COLUMN_NAME] = '%s'" % (table_name, geo_col)
                )

                rows = cursor.fetchall()
                if rows:

                    # Sample to parse:
                    # ([geom].[STGeometryType]()='Point')
                    type_re = re.compile(r"\[STGeometryType\]\(\)='(\w+)'")
                    for row in rows:
                        constraint = row[0]
                        m = type_re.search(constraint)
                        if m:
                            field_type = OGRGeomType(m.group(1)).django
                            break

        finally:
            cursor.close()

        return field_type, field_params



from decimal import Decimal

from django.contrib.gis.db.backends.base import BaseSpatialOperations
from django.contrib.gis.db.backends.util import SpatialFunction
from django.contrib.gis.geometry.backend import Geometry
from django.contrib.gis.measure import Distance
from django.utils import six
from sql_server.pyodbc.operations import DatabaseOperations


class MSSqlBoolMethod(SpatialFunction):
    """SQL Server (non-static) spatial functions are treated as methods,
    for eg g.STContains(p)"""

    sql_template = '%(geo_col)s.%(function)s(%(geometry)s) = 1'

    def __init__(self, function, **kwargs):
        super(MSSqlBoolMethod, self).__init__(function, **kwargs)


class MSSqlDistanceFunc(SpatialFunction):
    """Implements distance comparison lookups, eg distance_lte"""

    sql_template = ('%(geo_col)s.%(function)s(%(geometry)s) '
                    '%(operator)s %(result)s')

    def __init__(self, op):
        super(MSSqlDistanceFunc, self).__init__('STDistance',
                                                operator=op,
                                                result='%s')


class MSSqlBBBoolMethod(MSSqlBoolMethod):
    """SQL Server has no native bounding-box methods, but we can emulate
    them with a slightly more complicated expression.  The call will
    be translated into something like
    col.STEnvelope().STOverlaps(geom.STEnvelope())
    where STEnvelope() first simplifies the geometries to their
    bounding rectangles."""

    sql_template = '%(geo_col)s.STEnvelope().%(function)s(%(geometry)s.STEnvelope()) = 1'

    def __init__(self, function, **kwargs):
        super(MSSqlBoolMethod, self).__init__(function, **kwargs)


class MSSqlAdapter(str):
    """This adapter works around an apparent bug in the pyodbc driver
    itself.  We only require the wkt adapter, but if we use
    django.contrib.gis.db.backends.adapter.WKTAdapter then
    cursor.execute() fails because it doesn't call str() on unrecognised
    types.  So we make sure that our adaper *is* a string."""

    def __new__(cls, geom):
        geostr = str.__new__(cls, geom.wkt)
        geostr.srid = geom.srid
        return geostr

    def __eq__(self, other):
        if not isinstance(other, MSSqlAdapter):
            return False
        return super(MSSqlAdapter, self).__eq__(other) and \
            self.srid == other.srid

    def prepare_database_save(self, unused):
        return self


# Valid distance types and substitutions
dtypes = (Decimal, Distance, float) + six.integer_types


class MSSqlOperations(DatabaseOperations, BaseSpatialOperations):

    name = 'SQL Server'
    select = '%s.STAsText()'

    Adapter = MSSqlAdapter
    Adaptor = Adapter  # Backwards-compatibility alias.

    compiler_module = 'django_pyodbc_gis.compiler'

    geometry = True
    geography = True

    # Clearly a bald-faced lie, but the limitations of SQL Server are
    # actually similar to those of mysql, and coupled with some
    # baked-in mysql checks in the framework this lie enables the
    # admin to run.  See geometry_columns() and spatial_ref_sys().
    mysql = True

    # 'bbcontains'
    # 'bboverlaps'
    # 'contained'
    # 'contains'
    # 'contains_properly'
    # 'coveredby'
    # 'covers'
    # 'crosses'
    # 'disjoint'
    # 'distance_gt'
    # 'distance_gte'
    # 'distance_lt'
    # 'distance_lte'
    # 'dwithin'
    # 'equals'
    # 'exact'
    # 'intersects'
    # 'overlaps'
    # 'relate'
    # 'same_as'
    # 'touches'
    # 'within'
    # 'left'
    # 'right'
    # 'overlaps_left'
    # 'overlaps_right'
    # 'overlaps_above'
    # 'overlaps_below'
    # 'strictly_above'
    # 'strictly_below'

    geometry_functions = {
        'bbcontains': MSSqlBBBoolMethod('STContains'),
        'bboverlaps': MSSqlBBBoolMethod('STOverlaps'),
        'contained': MSSqlBBBoolMethod('STWithin'),
        'contains': MSSqlBoolMethod('STContains'),
        'crosses': MSSqlBoolMethod('STCrosses'),
        'disjoint': MSSqlBoolMethod('STDisjoint'),
        'equals': MSSqlBoolMethod('STEquals'),  # can we also implement exact, same_as like this?
        'intersects': MSSqlBoolMethod('STIntersects'),
        'overlaps': MSSqlBoolMethod('STOverlaps'),
        'touches': MSSqlBoolMethod('STTouches'),
        'within': MSSqlBoolMethod('STWithin'),
    }

    distance_functions = {
        'distance_gt': (MSSqlDistanceFunc('>'), dtypes),
        'distance_gte': (MSSqlDistanceFunc('>='), dtypes),
        'distance_lt': (MSSqlDistanceFunc('<'), dtypes),
        'distance_lte': (MSSqlDistanceFunc('<='), dtypes),
    }
    geometry_functions.update(distance_functions)

    geography_functions = {
        'contains': MSSqlBoolMethod('STContains'),
        'disjoint': MSSqlBoolMethod('STDisjoint'),
        'equals': MSSqlBoolMethod('STEquals'),
        'intersects': MSSqlBoolMethod('STIntersects'),
        'overlaps': MSSqlBoolMethod('STOverlaps'),
        'within': MSSqlBoolMethod('STWithin'),
    }
    geography_functions.update(distance_functions)

    gis_terms = set(geometry_functions) | set(['isnull'])

    collect = 'CollectionAggregate'
    extent = 'EnvelopeAggregate'
    unionagg = 'UnionAggregate'

    valid_aggregates = dict([(k, None) for k in
                             ('Collect', 'Extent', 'Union')])

    def spatial_lookup_sql(self, lvalue, lookup_type, value, field, qn):
        alias, col, db_type = lvalue

        geo_col = '%s.%s' % (qn(alias), qn(col))

        if field.geography:
            if lookup_type not in self.geography_functions:
                raise TypeError("Got invalid lookup_type for geography: %s" %
                                lookup_type)
        else:
            if lookup_type not in self.geometry_functions:
                raise TypeError("Got invalid lookup_type for geometry: %s" %
                                lookup_type)

        if lookup_type == 'isnull':
            return "%s IS %sNULL" % (geo_col, ('' if value else 'NOT ')), []

        else:
            if field.geography:
                op = self.geography_functions[lookup_type]
            else:
                op = self.geometry_functions[lookup_type]

            # if lookup_type is a tuple then we expect the value to be
            # a tuple as well:
            if isinstance(op, tuple):
                dist_op, arg_type = op

                # Ensuring that a tuple _value_ was passed in from the user
                if not isinstance(value, tuple):
                    raise ValueError('Tuple required for `%s` lookup type.' %
                                     lookup_type)
                if len(value) != 2:
                    raise ValueError('2-element tuple required for %s lookup type.' %
                                     lookup_type)

                # Ensuring the argument type matches what we expect.
                if not isinstance(value[1], arg_type):
                    raise ValueError('Argument type should be %s, got %s instead.' %
                                     (arg_type, type(value[1])))

                geom = value[0]
                return dist_op.as_sql(geo_col, self.get_geom_placeholder(field, geom))

            return op.as_sql(geo_col, self.get_geom_placeholder(field, value))

    def check_aggregate_support(self, aggregate):
        """
        Checks if the given aggregate name is supported (that is, if it's
        in `self.valid_aggregates`).
        """
        agg_name = aggregate.__class__.__name__
        return agg_name in self.valid_aggregates

    def spatial_aggregate_sql(self, agg):
        """
        Returns the spatial aggregate SQL template and function for the
        given Aggregate instance.
        """
        agg_name = agg.__class__.__name__
        if not self.check_aggregate_support(agg):
            raise NotImplementedError('%s spatial aggregate is not implmented '
                                      'for this backend.' % agg_name)
        agg_name = agg_name.lower()
        if agg_name == 'union':
            agg_name += 'agg'
        # We need to namespace the function depending on whether it's
        # for a geography or geometry (which requires digging into the
        # Aggregate), but the function name is the same for both:
        ns = 'geography' if agg.source.geography else 'geometry'
        sql_template = ns + '::%(function)s(%(field)s).ToString()'
        sql_function = getattr(self, agg_name)
        return sql_template, sql_function

    def convert_extent(self, poly):
        """
        Returns a 4-tuple extent for the `Extent` aggregate by converting
        the bounding box text returned by SQL Server (`poly` argument), for
        example: "POLYGON ((0 0, 2 0, 2 3, 0 3, 0 0))".
        """
        crnrs = poly[10:-2].split(',')
        xmin, ymin = map(float, crnrs[0].strip().split(' '))
        xmax, ymax = map(float, crnrs[2].strip().split(' '))
        return xmin, ymin, xmax, ymax

    def convert_geom(self, hex, geo_field):
        """
        Converts the geometry returned from aggregate queries.
        """
        if hex:
            return Geometry(hex)
        else:
            return None

    # GeometryField operations
    def geo_db_type(self, f):
        # We only have the one geometry type (especially since we
        # don't currently support geography):
        return 'geometry'

    def get_distance(self, f, value, lookup_type):
        """
        Returns the distance parameters for the given geometry field,
        lookup value, and lookup type.  This is based on the Spatialite
        backend, since we don't currently support geography operations.
        """
        if not value:
            return []
        value = value[0]
        if isinstance(value, Distance):
            if f.geodetic(self.connection):
                raise ValueError('The SQL Server backend does not support '
                                 'distance queries on geometry fields with '
                                 'a geodetic coordinate system. Distance '
                                 'objects; use a numeric value of your '
                                 'distance in degrees instead.')
            else:
                dist_param = getattr(value, Distance.unit_attname(f.units_name(self.connection)))
        else:
            dist_param = value
        return [dist_param]

    def get_geom_placeholder(self, f, value):
        """
        Because SQL Server does not support spatial transformations,
        there is no need to modify the placeholder based on the
        contents of the given value.  We do need to specify the SRID
        however, since this argument is required.
        """
        if hasattr(value, 'expression'):
            placeholder = self.get_expression_column(value)
        else:
            ns = 'geography' if f.geography else 'geometry'
            placeholder = '%s::STGeomFromText(%%s,%s)' % (ns, f.srid)
        return placeholder

    # Routines for getting the OGC-compliant models --- SQL Server
    # does not have OGC-compliant tables
    def geometry_columns(self):
        raise NotImplementedError

    def spatial_ref_sys(self):
        raise NotImplementedError
